from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.color.regex import ColorRegularExpression
from yta_general_utils.color.enums import ColorString
from colorsys import rgb_to_hsv as _rgb_to_hsv, hsv_to_rgb as _hsv_to_rgb
from typing import Union


def is_hexadecimal_color(color):
    """
    Check that the 'color' parameter is an hexadecimal
    color.
    """
    return ColorRegularExpression.HEX.parse(color)

def is_string_color(color):
    """
    Check that the 'color' parameter is an string 
    color accepted by our system, whose value is an
    hexadecimal value.
    """
    return ColorString.is_valid(color)

def is_array_or_tuple_without_alpha_normalized(color):
    """
    Check that the 'color' parameter is an array or a
    tuple of 3 elements that are float values between
    0 and 1 (normalized value).
    """
    return (
        is_array_or_tuple_without_alpha and
        all(
            PythonValidator.is_instance(c, float) and
            0 <= c <= 1
            for c in color
        )
    )

def is_array_or_tuple_with_alpha_normalized(color):
    """
    Check that the 'color' parameter is an array or a
    tuple of 4 elements that are float values between
    0 and 1 (normalized value).
    """
    return (
        is_array_or_tuple_with_alpha and
        all(
            PythonValidator.is_instance(c, float) and 
            0 <= c <= 1
            for c in color
        )
    )

def is_array_or_tuple_without_alpha(color):
    """
    Check that the 'color' parameter is an array or a
    tuple of 3 elements that are int values between 0
    and 255.
    """
    return (
        PythonValidator.is_instance(color, [tuple, list]) and
        len(color) == 3 and
        all(
            PythonValidator.is_instance(c, int) and
            0 <= c <= 255
            for c in color
        )
    )

def is_array_or_tuple_with_alpha(color):
    """
    Check that the 'color' parameter is an array or a
    tuple of 4 elements that are int values between 0
    and 255.
    """
    return (
        PythonValidator.is_instance(color, [tuple, list]) and
        len(color) == 4 and
        all(
            PythonValidator.is_instance(c, int) and
            0 <= c <= 255
            for c in color
        )
    )

def parse_rgb_color(color):
    """
    Parse the provided 'color' as RGB and returns it as
    r,g,b values.
    """
    if is_array_or_tuple_without_alpha_normalized(color):
        return color[0] * 255, color[1] * 255, color[2] * 255
    elif is_array_or_tuple_without_alpha(color):
        return color[0], color[1], color[2]
    else:
        raise Exception(f'The provided "color" parameter is not an RGB color.')

def parse_rgba_color(color):
    """
    Parse the provided 'color' as RGBA and returns it as
    r,g,b,a values.
    """
    if is_array_or_tuple_with_alpha_normalized(color):
        return color[0] * 255, color[1] * 255, color[2] * 255, color[3] * 255
    elif is_array_or_tuple_with_alpha(color):
        return color[0], color[1], color[2], color[3]
    else:
        raise Exception(f'The provided "color" parameter is not an RGBA color.')
    


def parse_color(color: Union[str, list, tuple]):
    """
    Tries to parse the provided 'color' and returns it
    as an RGBA if parseable, or None if not.
    """
    # As string color
    string_color = None
    try:
        string_color = ColorString.to_enum(color)
    except:
        pass

    color_array = None
    if string_color is not None:
        color_array = hex_to_rgba(string_color.value)
    # A hexadecimal string
    elif PythonValidator.is_string(color) and is_hexadecimal_color(color):
        color_array = hex_to_rgba(color)
    else:
        # RGBA color
        try:
            color_array = parse_rgba_color(color)
        except:
            pass

        # RGB color
        try:
            color_array = *parse_rgb_color(color), 0
        except:
            pass

    # TODO: What about HSL, CYMK, etc. (?)

    return color_array
    
# These methods below are just actioners, they don't
# check anything as they will be used by a class that
# validates everything before using these methods.
def hex_to_rgba(hex_color: str):
    # Hex can start with '0x', '0X' or '#'
    hex = hex_color.lstrip('#').lstrip('0x').lstrip('0X')
    if len(hex) == 8:
        # hex with alpha
        r, g, b, a = (int(hex[i:i+2], 16) for i in (0, 2, 4, 6))
    elif len(hex) == 6:
        # hex without alpha
        r, g, b, a = *(int(hex[i:i+2], 16) for i in (0, 2, 4)), 0
    
    return r, g, b, a

def hex_to_rgb(hex_color: str):
    r, g, b, _ = hex_to_rgba(hex_color)

    return r, g, b

def rgb_to_hex(rgb_color: list, do_include_alpha: bool = False):
    r, g, b = parse_rgb_color(rgb_color)

    return rgba_to_hex([r, g, b, 255], do_include_alpha)

def rgba_to_hex(rgba_color: list, do_include_alpha: bool = False):
    r, g, b, a = parse_rgba_color(rgba_color)

    hex = "#{:02x}{:02x}{:02x}".format(r, g, b)
    if do_include_alpha:
        hex = "#{:02x}{:02x}{:02x}{:02x}".format(r, g, b, a)

    return hex

def rgba_to_hsl(rgba_color: list):
    r, g, b, _ = parse_rgba_color(rgba_color)

    return rgb_to_hsl([r, g, b])

def rgb_to_hsl(rgb_color: list):
    r, g, b = parse_rgb_color(rgb_color)

    # Values normalization
    r /= 255.0
    g /= 255.0
    b /= 255.0
    
    # Max and minimum values for RGB
    cmax = max(r, g, b)
    cmin = min(r, g, b)
    delta = cmax - cmin

    # Tone (H)
    if delta == 0:
        h = 0  # No difference => undefined (gray) tone
    elif cmax == r:
        h = (60 * ((g - b) / delta) + 360) % 360
    elif cmax == g:
        h = (60 * ((b - r) / delta) + 120) % 360
    else:  # cmax == b
        h = (60 * ((r - g) / delta) + 240) % 360
    
    # Luminosity (L)
    l = (cmax + cmin) / 2
    
    # Saturation (S)
    if delta == 0:
        s = 0  # No difference => saturation is 0
    else:
        s = delta / (1 - abs(2 * l - 1)) if l != 0 and l != 1 else delta / (2 - (cmax + cmin))

    # TODO: I saw in some online solutions that they offer
    # the results without decimal figures
    return round(h, 2), round(s * 100, 2), round(l * 100, 2)

# TODO: Add 'hsl_to_rgb'
# TODO: Add 'hsl_to_rgba'

def rgb_to_cymk(rgb_color: list):
    r, g, b = parse_rgb_color(rgb_color)

    r, g, b = r / 255.0, g / 255.0, b / 255.0

    k = 1 - max(r, g, b)

    if k == 1:
        c = m = y = 0
    else:
        c = (1 - r - k) / (1 - k)
        m = (1 - g - k) / (1 - k)
        y = (1 - b - k) / (1 - k)

    # TODO: I saw in some online solutions that they offer
    # the results without decimal figures
    return round(c * 100, 2), round(m * 100, 2), round(y * 100, 2), round(k * 100, 2)

def rgb_to_hsv(rgb_color):
    r, g, b = parse_rgb_color(rgb_color)

    # TODO: Assume this is not normalized
    return _rgb_to_hsv(r, g, b)

def rgba_to_hsv(rgba_color):
    r, g, b, _ = parse_rgba_color(rgba_color)
    
    # TODO: Assume this is not normalized
    return _rgb_to_hsv(r, g, b)

def hsv_to_rgb(h, s, v):
    # TODO: Assume this is not normalized
    return _hsv_to_rgb(h, s, v)

def hsv_to_rgba(h, s, v):
    # TODO: Assume this is not normalized
    return *hsv_to_rgb(h, s, v), 255





from yta_general_utils.file.checker import FileValidator
from yta_general_utils.programming.parameter_validator import PythonValidator

import json


class FileReader:
    """
    Class to simplify and encapsulate the functionality related
    with reading files.
    """
    @staticmethod
    def read_json(filename: str):
        """
        Reads the provided 'filename' and returns the information 
        as a json (if possible).

        Parameters
        ----------
        filename : str
            File path from which we want to read the information.
        """
        if not PythonValidator.is_string(filename) or not FileValidator.file_exists(filename):
            raise Exception('The provided "filename" is not a valid string or filename.')
        
        with open(filename, encoding = 'utf-8') as json_file:
            return json.load(json_file)
        
    @staticmethod
    def read_lines(filename: str):
        """
        Read the content of the provided 'filename'
        if valid and return it as it decomposed in
        lines.

        Parameters
        ----------
        filename : str
            File path from which we want to read the information.
        """
        if not PythonValidator.is_string(filename) or not FileValidator.file_exists(filename):
            raise Exception('The provided "filename" is not a valid string or filename.')
        
        with open(filename, 'r', encoding = 'utf-8') as file:
            return file.readlines()
        
    @staticmethod
    def read(filename: str):
        """
        Read the content of the provided 'filename'
        if valid and return it as it is.

        Parameters
        ----------
        filename : str
            File path from which we want to read the information.
        """
        if not PythonValidator.is_string(filename) or not FileValidator.file_exists(filename):
            raise Exception('The provided "filename" is not a valid string or filename.')
        
        with open(filename, 'r', encoding = 'utf-8') as file:
            return file.read()



from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from yta_general_utils.random import random_int_between
from yta_general_utils.math import Math

import numpy as np


NORMALIZATION_MIN_VALUE = -10000
"""
The lower limit for the normalization process.
"""
NORMALIZATION_MAX_VALUE = 10000
"""
The upper limit for the normalization process.
"""

class CoordinateType(Enum):
    """
    Enum class to represent a coordinate type that 
    defines the way to make the calculations.
    """
    CORNER = 'corner'
    """
    The type of coordinate that represents the upper
    left corner of the object that uses the coordinate
    with this type.
    """
    CENTER = 'center'
    """
    The type of coordinate that represents the center of
    the object that uses the coordinate with this type.
    """

class Coordinate:
    """
    Class to represent a coordinate point ('x', 'y').
    """
    position: tuple = None
    """
    The ('x', 'y') tuple containing the position
    coordinate.
    """
    _is_normalized: bool = False
    """
    Internal function to know if it has been normalized
    or not.
    """

    @property
    def x(self):
        return self.position[0]
    
    @property
    def y(self):
        return self.position[1]
    
    @property
    def is_normalized(self):
        return self._is_normalized

    def __init__(self, x: float, y: float, is_normalized: bool = False):
        if not NumberValidator.is_number_between(x, NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE) or not NumberValidator.is_number_between(y, NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE):
            raise Exception(f'The "x" and "y" parameters must be values between {str(NORMALIZATION_MIN_VALUE)} and {str(NORMALIZATION_MAX_VALUE)} and "{str(x)}, {str(y)}" provided.')
        
        if not PythonValidator.is_boolean(is_normalized):
            raise Exception('The "is_normalized" parameter must be a boolean value.')
        
        self.position = (x, y)
        self._is_normalized = is_normalized

    def get_x(self):
        """
        Return the 'x' value.
        """
        return self.x
    
    def get_y(self):
        """
        Return the 'y' value.
        """
        return self.y

    def as_tuple(self):
        """
        Return the coordinate as a tuple ('x', 'y').
        """
        return Coordinate.to_tuple(self)
    
    def as_array(self):
        """
        Return the coordinate as an array ['x', 'y'].
        """
        return Coordinate.to_array(self)

    def normalize(self):
        """
        Normalize the coordinate by turning the values into
        a range between [0.0, 1.0]. This will be done if the
        values have not been normalized previously.
        """
        if not self._is_normalized:
            self.position = Coordinate.normalize_tuple(self.position)
            self._is_normalized = True

        return self

    def denormalize(self):
        """
        Denormalize the coordinate values by turning them
        from normalized values to the real ones. This will
        be done if the values have been normalized 
        previously.
        """
        if self._is_normalized:
            self.position = Coordinate.denormalize_tuple(self.position)
            self._is_normalized = False

        return self

    @staticmethod
    def to_tuple(coordinate):
        """
        Turn the provided 'coordinate' to a tuple like ('x', 'y').
        """
        return coordinate.position
    
    @staticmethod
    def to_array(coordinate):
        """
        Turn the provided 'coordinate' to an array like ['x', 'y'].
        """
        return [coordinate.x, coordinate.y]

    @staticmethod
    def generate(amount: int = 1):
        """
        Generate 'amount' coordinates with random values
        between [0, 1920] for the 'x' and [0, 1080] for
        the 'y', that are returned as an array of instances.

        The 'amount' parameter is limited to the interval 
        [1, 100].
        """
        if not NumberValidator.is_number_between(amount, 1, 100):
            raise Exception(f'The provided "amount" parameter "{str(amount)}" is not a number between 1 and 100.')
        
        return Coordinate(random_int_between(0, 1920), random_int_between(0, 1080))
    
    @staticmethod
    def to_numpy(coordinates: list['Coordinate']):
        """
        Convert a list of Coordinates 'coordinates' to
        numpy array to be able to work with them.

        This method does the next operation:
        np.array([[coord.x, coord.y] for coord in coordinates])
        """
        if not PythonValidator.is_list(coordinates):
            if not PythonValidator.is_instance(coordinates, Coordinate):
                raise Exception('The provided "coordinates" parameter is not a list of NormalizedCoordinates nor a single NormalizedCoordinate instance.')
            else:
                coordinates = [coordinates]
        elif any(not PythonValidator.is_instance(coordinate, Coordinate) for coordinate in coordinates):
            raise Exception('At least one of the provided "coordinates" is not a NormalizedCoordinate instance.')

        return np.array([coordinate.as_array() for coordinate in coordinates])
    
    @staticmethod
    def normalize_tuple(coordinate: tuple):
        """
        Normalize the provided 'coordinate' by applying
        our normalization limits. This means turning the
        non-normalized 'coordinate' to a normalized one
        (values between 0.0 and 1.0).
        """
        return (
            Math.normalize(coordinate[0], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE),
            Math.normalize(coordinate[1], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE)
        )
    
    @staticmethod
    def denormalize_tuple(coordinate: tuple):
        """
        Denormalize the provided 'coordinate' by applying
        our normalization limits. This means turning the 
        normalized 'coordinate' (values between 0.0 and
        1.0) to the not-normalized ones according to our
        normalization limits.
        """
        return (
            Math.denormalize(coordinate[0], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE),
            Math.denormalize(coordinate[1], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE)
        )
    
    @staticmethod
    def is_valid(coordinate: tuple):
        """
        Check if the provided 'coordinate' is valid or not.
        A valid coordinate is a tuple with two elements that
        are values between our normalization limits.
        """
        if not PythonValidator.is_instance(coordinate, 'Coordinate') and (not PythonValidator.is_tuple(coordinate) or len(coordinate) != 2 or not NumberValidator.is_number_between(coordinate[0], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE) or not NumberValidator.is_number_between(coordinate[1], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE)):
            return False
        
        return True

    @staticmethod
    def validate(coordinate: tuple, parameter_name: str):
        """
        Validate if the provided 'coordinate' is a coordinate
        with values between our normalization limits.
        """
        if not Coordinate.is_valid(coordinate):
            raise Exception(f'The provided "{parameter_name}" parameter is not a valid tuple of 2 elements that are values between our limits {str(NORMALIZATION_MIN_VALUE)} and {str(NORMALIZATION_MAX_VALUE)}. Please, provide a valid coordinate.')


from yta_general_utils.coordinate.coordinate import Coordinate


class Region:
    """
    Class to represent a region built by two coordinates, one in
    the top left corner and another one in the bottom right 
    corner.
    """
    top_left: Coordinate = None
    bottom_right: Coordinate = None
    # TODO: Do we actually need them (?)
    coordinates: list[Coordinate] = None
    _width: int = None
    _height: int = None

    def __init__(self, top_left_x: int, top_left_y: int, bottom_right_x: int, bottom_right_y: int, coordinates: list[Coordinate]):
        self.top_left = Coordinate(top_left_x, top_left_y)
        self.bottom_right = Coordinate(bottom_right_x, bottom_right_y)
        # TODO: Do we actually need them (?)
        self.coordinates = coordinates
        self._width = self.bottom_right.x - self.top_left.x
        self._height = self.bottom_right.y - self.top_left.y

    @property
    def width(self):
        return self._width
    
    @property
    def height(self):
        return self._height
    
    @property
    def center(self):
        """
        The center position of the region represented by a tuple
        (x, y).
        """
        return self.bottom_right.x - self.width / 2, self.bottom_right.y - self.height / 2
    



from yta_general_utils.downloader.google_drive import GoogleDriveResource
from yta_general_utils.programming.regular_expressions import GeneralRegularExpression
from youtubeenhanced.objects.youtube_video import YoutubeVideo

import requests
import mimetypes


def url_is_ok(url: str):
    """
    Checks if the provided url is valid. It returns True if yes or
    False if not. This method uses a head request to check the 
    status_code of the response.
    """
    if not url:
        return False

    if not isinstance(url, str):
        raise Exception('The "url" parameter provided is not a str.')
    
    try:
        response = requests.head(url)

        if response.status_code == 200:
            return True
        else:
            return False
    except requests.ConnectionError as e:
        print('Something went wrong with "' + url + '" (exception raised).')
        return False
    
def verify_image_url(url: str):
    """
    This method will check that the provided 'url' is a valid one, that
    it is an image, and also will return the image file extension (if
    available), or False if not valid 'url'. The extension returned, if
    returned, contains the dot '.'.

    Use this method before trying to download an image from a url.

    This method will send a head request to the provided 'url', check
    the status_code and also the content-type.
    """
    if not url:
        # TODO: Maybe raise Exception (?) I think no...
        return False
    
    try:
        response = requests.head(url)

        if response.status_code != 200: # Maybe more than only 200 are valid
            return False
        
        if not response.headers['content-type'].startswith('image/'):
            return False
        
        # This below is like 'image/jpeg', 'application/pdf' so maybe we can just
        # split by '/' and obtain the second part, because when I guess from a
        # 'image/jpeg' content-type, the response is '.jpg' and not '.jpeg'...
        content_type = response.headers['content-type']
        extension = mimetypes.guess_extension(content_type)

        # TODO: Review this below when tested 
        # This 'other_extension' below could be a better choice maybe (read above)
        other_extension = '.' + content_type.split('/')[1]

        return extension
    except requests.ConnectionError as e:
        print('Something went wrong with "' + url + '" (exception raised).')
        return False
    
def is_google_drive_url(google_drive_url: str):
    """
    Checks if the provided 'google_drive_url' is a string with a
    valid Google Drive url format or not. It returns True if yes
    or False if not.
    """
    try:
        GoogleDriveResource(google_drive_url)
    except:
        return False
    
    return True

def is_youtube_video_url(youtube_video_url: str):
    """
    Checks if the provided 'youtube_video_url' is string containing
    a valid Youtube video url. This method only checks that the
    given url is well formatted.

    Example of url : https://www.youtube.com/watch?v=OpA2ZxnRs6
    """
    return GeneralRegularExpression.YOUTUBE_VIDEO_URL.parse(youtube_video_url)

def is_valid_youtube_video(youtube_video_url: str):
    """
    Checks if the provided 'youtube_video_url' is a valid Youtube
    video url and also that the video exists and is accessible (this
    is done by checking that duration is not None).

    Example of url : https://www.youtube.com/watch?v=OpA2ZxnRs6
    """
    if not is_youtube_video_url(youtube_video_url):
        return False
    
    # TODO: If this works, use the 'is_valid' method instead I
    # created in YoutubeVideo class. If not, fix it also there
    return YoutubeVideo(youtube_video_url).duration is not None


import inspect


# TODO: This whole file has to be reviewed and refactored
# or maybe removed due to new PythonValidator, NumberValidator
# etc.
def code_file_is(parameter, filename):
    """
    Checks if the provided 'parameter' code is contained in the also
    provided 'filename'. This method is useful to check Enum objects
    or classes as we know the name we use for the files.

    This method was created to be able to check if a function that
    was passed as parameter was part of a custom Enum we created
    and so we could validate the was actually that custom Enum.
    Working with classes was not returning the custom Enum class
    created, so we needed this.
    """
    # TODO: Is this useful or needed (?)
    if inspect.getfile(parameter).endswith(filename):
        return True
    
    return False


#from yta_general_utils.programming.parameter_validator import ParameterValidator
from yta_general_utils.programming.output import handle_output_filename
from yta_general_utils.downloader.audio import download_audio
from yta_general_utils.downloader.gif import download_gif
from yta_general_utils.downloader.google_drive import GoogleDriveResource
from yta_general_utils.downloader.image import download_image_2
from yta_general_utils.downloader.video import download_video


# TODO: Make all submethods return the element read (image as Pillow,
# video as VideoFileClip or numpy, etc. and also the filename
class Downloader:
    """
    Class to encapsulate the functionality related to download resources
    from the Internet.
    """
    # TODO: Maybe move the checkings to the specific 'download_x' method
    # and not here that is more a encapsulation class
    @staticmethod
    def download_audio(url: str, output_filename: str):
        """
        Download the audio file from the provided 'url' and stores
        it locally as 'output_filename'.
        """
        # TODO: Refactor the parameter validation to avoid cyclic imports
        #ParameterValidator.validate_string_mandatory_parameter(url)

        # TODO: Add more checkings here (?)
        output_filename = handle_output_filename(output_filename)

        return download_audio(url, output_filename)
    
    @staticmethod
    def download_gif(query: str, output_filename: str):
        """
        Search for a gif with the provided 'query' and download it,
        if existing, to a local file called 'output_filename'.
        """
        #ParameterValidator.validate_string_mandatory_parameter(query)
        output_filename = handle_output_filename(output_filename, '.gif')

        return download_gif(query, output_filename)
    
    @staticmethod
    def download_google_drive_resource(google_drive_url: str, output_filename: str):
        """
        Download the Google Drive resource from the given 'google_drive_url',
        if existing and available, to a local file called 'output_filename'.
        """
        #ParameterValidator.validate_string_mandatory_parameter(google_drive_url)
        output_filename = handle_output_filename(output_filename)
        
        return GoogleDriveResource.download_from_url(google_drive_url, output_filename)
    
    @staticmethod
    def download_image(url: str, output_filename: str):
        """
        Download the image from the provided 'url' and stores it, if
        existing and available, as a local file called 'output_filename'.
        """
        #ParameterValidator.validate_string_mandatory_parameter(url)
        output_filename = handle_output_filename(output_filename)
        
        return download_image_2(url, output_filename)

    @staticmethod
    def download_video(url: str, output_filename: str):
        """
        Download the video from the provided 'url' and stores it, if
        existing and available, as a local file called 'output_filename'.
        """
        #ParameterValidator.validate_string_mandatory_parameter(url)
        output_filename = handle_output_filename(output_filename)
        
        return download_video(url, output_filename)





from bing_image_downloader.bing import Bing
from google_images_download import google_images_download

import re
import urllib


def test(keywords, images_number, output_folder):
    HEADERS = {
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) ' 
        'AppleWebKit/537.11 (KHTML, like Gecko) '
        'Chrome/23.0.1271.64 Safari/537.11',
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
        'Accept-Charset': 'ISO-8859-1,utf-8;q=0.7,*;q=0.3',
        'Accept-Encoding': 'none',
        'Accept-Language': 'en-US,en;q=0.8',
        'Connection': 'keep-alive'
    }
    #gid = google_images_download()
    #gid.download(keywords = keywords, size = 'large', aspect_ratio = 'panoramic')
    
    #bing = Bing(keywords, images_number, output_folder, False, 100, False, False)
    # Parse the page source and download pics
    request_url = 'https://www.bing.com/images/async?q=' + urllib.parse.quote_plus(keywords) \
                    + '&first=' + '0' + '&count=' + str(images_number) \
                    + '&adlt=' + 'False' + '&qft=' + 'False'
    request = urllib.request.Request(request_url, None, headers = HEADERS)
    response = urllib.request.urlopen(request)
    html = response.read().decode('utf8')
    if html ==  "":
        return None
    links = re.findall('murl&quot;:&quot;(.*?)&quot;', html)
    for link in links:
        print(link)
        #if self.download_count < self.limit and link not in self.seen:
            #self.seen.add(link)
            #self.download_image(link)


    # download(keywords, limit = images_number, output_dir = output_folder)



# Based on this: https://stackoverflow.com/a/5191224
class ClassPropertyDescriptor(object):
    def __init__(self, fget, fset = None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, cls = None):
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self

def classproperty(func):
    """
    Decorator to implement a class property.
    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


from yta_general_utils.file.remover import FileRemover
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.programming.env import get_current_project_env
from random import randint
from datetime import datetime

import os


# TODO: Turn all this into a TempFileHandler class (?)
def initialize():
    """
    This method loads the current project dotenv ('.env' file in
    root folder), looks for a 'WIP_FOLDER' variable definition 
    in that file and, if not defined, sets it as the default
    'yta_wip' value and creatis the folder if it doesn't exist.

    This method is to ensure that there is a temporary folder to
    work with and available since the begining.
    """
    WIP_FOLDER = get_current_project_env('WIP_FOLDER')
    if not WIP_FOLDER:
        # We force creating the dir
        WIP_FOLDER = get_project_abspath() + 'yta_wip/'
        if not FileValidator.is_folder(WIP_FOLDER):
            os.mkdir(WIP_FOLDER)

    return WIP_FOLDER

WIP_FOLDER = initialize()

def get_temp_filename(filename):
    """
    Receives a 'filename' and turns it into a temporary filename that is
    returned including a random suffix datetime related.

    This method uses the current datetime and a random integer number to
    be unique.

    If you provide 'file.wav' it will return something like 
    'file_202406212425.wav'.
    """
    delta = (datetime.now() - datetime(1970, 1, 1))
    aux = filename.split('.')

    return aux[0] + '_' + str(int(delta.total_seconds())) + str(randint(0, 10000)) + '.' + aux[1]

def create_temp_filename(filename):
    """
    Returns a temporary file name that includes the 'WIP_FOLDER'
    set in environment variable and also a random and datetime
    related suffix.

    The WIP (Work In Progress) folder will be set to 'yta_wip'
    if not found in .env.

    If you provide 'file.wav' it will return something like 
    '$WIP/file_202406212425.wav'.
    """
    # TODO: Rename this as it uses wip and we do not mention it
    # TODO: Issue if no extension provided
    return create_custom_temp_filename(get_temp_filename(filename))

# TODO: Maybe rename this methods...
def create_custom_temp_filename(filename):
    """
    Returns a new 'filename' that includes the 'WIP_FOLDER' but
    preserves the original name. This is for using the temporary
    folder but without any internal logic.

    The WIP (Work In Progress) folder will be set to 'yta_wip'
    if not found in .env.
    """
    return WIP_FOLDER + filename

def clean_temp_folder():
    """
    Removes all the existing files in the temporary folder. This folder is
    the one set in the environment 'WIP_FOLDER' variable.

    The WIP (Work In Progress) folder will be set to 'yta_wip'
    if not found in .env.
    """
    FileRemover.delete_files(WIP_FOLDER)


from PIL import Image
from io import BytesIO

import numpy as np
import base64
import cv2


class ImageConverter:
    @classmethod
    def validate_numpy_image(cls, image: np.ndarray):
        """
        This method checks if the provided 'image' is a numpy array, that the
        array has 3 or 4 elements in each cell and if its values are in the
        [0, 255] range or in the [0, 1] (normalized) range. It will raise an
        Exception if any of those conditions are not satisfied.
        """
        if not isinstance(image, np.ndarray):
            raise Exception('The provided "image" parameter is not a numpy np.ndarray instance.')
        
        if image.ndim != 3 or image.shape[2] not in [3, 4]:
            raise Exception('The provided "image" parameter does not represent a RGB or RGBA image.')

        if not np.all((image >= 0) & (image <= 255)):
            raise Exception('The provided numpy array is not a valid image as its values are not between 0 and 255.')

        # TODO: What about '.astype('uint8')', maybe we can check if it is that type (?)

    @classmethod
    def validate_opencv_image(cls, image: np.ndarray):
        """
        This method checks if the provided 'image' is a numpy array, that the
        array has 3 or 4 elements in each cell and if its values are in the
        [0, 255] range or in the [0, 1] (normalized) range. It will raise an
        Exception if any of those conditions are not satisfied.

        An opencv image is just a numpy array with some meta param.
        """
        # The only thing that could change is the message and I don't want
        # to duplicate code for a single 'opencv' word in a message
        return cls.validate_numpy_image(image)

    @classmethod
    def validate_base64_image(cls, image: str):
        """
        This method validates if the provided image is a valid base64 image
        by getting the prefix, the 'base64' str and also trying to decode it.
        It will raise an Exception if the image is not a valid base64 image.
        """
        is_valid = False

        if image.startswith('data:image/') and ';base64,' in image:
            base64_string = image.split(';base64,')[1]
            try:
                base64.b64decode(base64_string, validate = True)
                is_valid = True
            except Exception:
                pass
            
        if not is_valid:
            raise Exception('The provided "image" parameter is not a valid base64 image.')

    @classmethod
    def validate_pillow_image(cls, image: Image.Image):
        if not isinstance(image, Image.Image):
            raise Exception('The provided "image" is not an instance of a Pillow image.')
        
        if image.mode not in ['RGB', 'RGBA']:
            raise Exception('The provided pillow image is not in a valid mode for our software. Valid modes are: "RGB", "RGBA".')
        
    @classmethod
    def numpy_image_to_pil(cls, image: np.ndarray):
        """
        This method checks if the provided 'image' is a numpy array and if its
        values are in the [0, 255] range or in the [0, 1] (normalized) range.
        It will raise an Exception if any of those conditions are not
        satisfied.

        This method will return the image converted into a Pillow image.
        """
        cls.validate_numpy_image(image)
        
        if np.all((image >= 0) & (image <= 1)):
            # TODO: How do I know if the values are normalized or just [0, 255]
            # but with values below 1 (?)
            image = Image.fromarray((image).astype(np.uint8))
        else:
            image = Image.fromarray((image * 255).astype(np.uint8))

        return image
    
    @classmethod
    def numpy_image_to_base64(cls, image: np.ndarray):
        """
        Turns the provided numpy 'image' into a base64 str image.
        """
        cls.validate_numpy_image(image)

        buffer = BytesIO()
        image = cls.numpy_image_to_pil(image).save(buffer, format = 'PNG')
        buffer.seek(0)
        image_bytes = buffer.read()

        image_base64 = base64.b64encode(image_bytes).decode('utf-8')
        image_base64_str = f'data:image/png;base64,{image_base64}'

        return image_base64_str
    
    @classmethod
    def numpy_image_to_opencv(cls, image: np.ndarray):
        cls.validate_numpy_image(image)
    
        # This is also a way:
        # pil_data = PIL.Image.open('Image.jpg').convert('RGB')
        # image = numpy.array(pil_data)[:, :, ::-1].copy()

        # I need to know if image is RGB or RGBA
        if image.ndim == 3 and image.shape[2] == 3:   # RGB
            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        elif image.ndim == 2 and image.shape[2] == 4:   # RGBA
            image = cv2.cvtColor(image, cv2.COLOR_RGBA2BGRA)
        
        return image
    
    @classmethod
    def pil_image_to_numpy(cls, image: Image.Image):
        """
        Turns the 'image' to a numpy array. The PIL image must be an
        array produced by the code 'Image.open(image_filename)'.
        """
        cls.validate_pillow_image(image)

        # This will return it as RGB if (4, 4, 3) or as RGBA if (4, 4, 4)
        return np.asarray(image)
    
    @classmethod
    def pil_image_to_base64(cls, image: Image.Image):
        """
        Turns the 'image' to a base64 image by turning it into a numpy
        image first. The PIL image must be an array produced by the code
        'Image.open(image_filename)'.
        """
        cls.validate_pillow_image(image)

        return cls.numpy_image_to_base64(cls.pil_image_to_numpy(image))

    @classmethod
    def pil_image_to_opencv(cls, image: Image.Image):
        """
        Turns the 'image' to a opencv image by turning it into a numpy
        image first. The PIL image must be an array produced by the code
        'Image.open(image_filename)'.
        """
        cls.validate_pillow_image(image)

        return cls.numpy_image_to_opencv(cls.pil_image_to_numpy(image))
    
    @classmethod
    def base64_image_to_pil(cls, image):
        """
        Turns the 'image' to a PIL Image, to be able
        to work with, and returns it.
        """
        cls.validate_base64_image(image)

        return Image.open(BytesIO(base64.b64decode(image)))
    
    @classmethod
    def base64_image_to_numpy(cls, image):
        """
        Turns the 'image' to a numpy image (np.ndarray),
        to be able to work with, and returns it. 
        """
        cls.validate_base64_image(image)
        
        return cls.pil_image_to_numpy(cls.base64_image_to_pil(image))
    
    @classmethod
    def base64_image_to_opencv(cls, image):
        """
        Turns the 'image' to an opencv image by turning it
        into a numpy array first.
        """
        cls.validate_base64_image(image)

        return cls.pil_image_to_base64(cls.base64_image_to_pil)

    @classmethod
    def opencv_image_to_numpy(cls, image: np.ndarray):
        """
        Turns the 'image' to an opencv image by turning it
        into a numpy array first.
        """
        cls.validate_opencv_image(image)

        # An opencv image is just a numpy array with a meta param

        # I need to know if image is RGB or RGBA
        if image.ndim == 3 and image.shape[2] == 3:   # RGB
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        elif image.ndim == 2 and image.shape[2] == 4:   # RGBA
            image = cv2.cvtColor(image, cv2.COLOR_BGRA2RGBA)
        
        return image

    @classmethod
    def opencv_image_to_pillow(cls, image: np.ndarray):
        """
        Turns the 'image' to a pillow image by turning it
        into a numpy array first.
        """
        cls.validate_opencv_image(image)

        return cls.numpy_image_to_pil(cls.opencv_image_to_numpy(image))
    
    @classmethod
    def opencv_image_to_base64(cls, image: np.ndarray):
        """
        Turns the 'image' to a base64 image by turning it
        into a numpy array first.
        """
        cls.validate_opencv_image(image)

        return cls.numpy_image_to_base64(cls.opencv_image_to_numpy(image))

